/*
** Module   : LabeledBorderedPanel.java
** Abstract : Panel with border and (optionally) - side or top label.
**
** Copyright (c) 2010-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 SIY 20108020          Refactored common code from RadioSet and
**                           SelectionList.
** 002 SIY 20111101          Moved from CHUI package. Minor 
**                           cleanups/refactorings in preparation to GUI client.
** 003 SIY 20111202          Updated imports.
** 004 EVL 20140528          Adding implementation stubs for two new methods.
*/

package com.goldencode.p2j.ui.client;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.border.*;
import com.goldencode.p2j.ui.client.layout.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Common implementation of {@link LabeledWidget} interface for 
 * {@link RadioSet} and {@link SelectionList}.
 * 
 * @author 	SIY
 * @version 1.0
 */
public class LabeledBorderedPanel 
extends BorderedPanel
implements LabeledWidget
{
   /** Component label. */
   private Label label = null;
   
   /**
    * Default constructor.
    */
   public LabeledBorderedPanel()
   {
   }
   
   /**
    * Constructor.
    * 
    * @param   mgr
    *          Layout manager.
    * @param   border
    *          Border for this widget.
    */
   public LabeledBorderedPanel(LayoutManager mgr, Border border)
   {
      super(mgr, border);
   }

   /**
    * Provide access to the attributes stored in the widget.
    * 
    * @return  Reference to attribute container.
    */
   @SuppressWarnings("unchecked")
   @Override
   public ControlSetConfig config()
   {
      return null;
   }
   
   /**
    * Return current value of COLON option.
    * 
    * @return  Current value of the option column.
    */
   @Override
   public int getColon()
   {
      return config().getColon();
   }
   
   /**
    * Return text of the column label.
    * 
    * @return  text of the column label.
    */
   @Override
   public String getColumnLabelText()
   {
      return config().getColumnLabel();
   }
   
   /**
    * Return reference to the instance of the <code>Label</code> attached to
    * the widget.
    * 
    * @return  Reference to <code>Label</code> instance.
    */
   @Override
   public Label getLabelInstance()
   {
      return label;
   }

   /**
    * Return text of the label.
    * 
    * @return  text of the label.
    */
   @Override
   public String getLabelText()
   {
      return config().getLabel();
   }

   /**
    * Return TO column.
    * 
    * @return  Column assigned via TO option.
    */
   @Override
   public int getTo()
   {
      return config().getTo();
   }

   /**
    * Check if widget has fixed column position.
    * 
    * @return  <code>true</code> if widget column is fixed.
    */
   @Override
   public boolean isColumnFixed()
   {
      return config().getAssignedColumn() != BaseConfig.INV_COORD;
   }

   /**
    * Get current state of flag which is forces presence of the label
    * regardless from the frame NO-LABELS option.
    * 
    * @return  Current state of the flag.
    */
   @Override
   public boolean isForceLabel()
   {
      return config().isForceLabel();
   }
   
   /**
    * Get current state of NO-LABELS flag.
    * 
    * @return  Current state of the flag.
    */
   @Override
   public boolean isNoLabels()
   {
      return config().isNoLabels();
   }
   
   /**
    * Get widget alignment policy.
    * 
    * @return  <code>true</code> if widget is formatted for right alignment.
    */
   @Override
   public boolean isRightAligned()
   {
      return false;
   }

   /**
    * Set new value for the COLON option.
    * 
    * @param   colon
    *          New value for the option column.
    */
   @Override
   public void setColon(int colon)
   {
      config().setColon(colon);
   }

   /**
    * Assign new instance of <code>Label</code> to the widget.
    * 
    * @param   label
    *          Reference to new <code>Label</code> instance.
    */
   @Override
   public void setLabelInstance(Label label)
   {
      this.label = label;
      this.label.setVisible(isVisible());
   }

   /**
    * Force NO-LABELS flag.
    * 
    * @param   labels
    *          New value of the flag.
    */
   @Override
   public void setNoLabels(boolean labels)
   {
      config().setNoLabels(labels);
   }

   /**
    * Assign new TO column.
    * 
    * @param   col
    *          New value for the option.
    */
   @Override
   public void setTo(int col)
   {
      config().setTo(col);
   }


   /**
    * Shows or hides the radio-set and its label.
    *
    * @param    visible
    *           <code>true</code> to make the widget visible, 
    *           <code>false</code> to hide the widget.
    */
   @Override
   public void setVisible(boolean visible)
   {
      super.setVisible(visible);

      if (label != null)
         label.setVisible(visible);
   }
   
   /**
    * Checking if the widget was made visible at least once.
    * 
    * @return  <code>true</code> if the widget was visible,
    *          <code>false</code> otherwise.
    */
   @Override
   public boolean wasVisible()
   {
      return true;
   }
   
   /**
    * Remember the widget was made visible at least once.
    */
   @Override
   public void setWasVisible()
   {
      return;
   }
}
