/*
** Module   : AbstractContainer.java
** Abstract : Base class for widget containers.
**
** Copyright (c) 2010-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -----------------------Description------------------------
** 001 SIY 20100629 Created initial version
** 002 SIY 20100804 Changes in imports, some refactorings.
** 003 SIY 20100831 Changes related to refactoring of FrameHelper.
** 004 SIY 20101122 Switching from KeyEvent to KeyInput.
** 005 SIY 20110425 Fixed next/prev focus handling for deeply nested 
**                  containers.
** 006 GES 20111012 Moved static data to context-local. This allows multiple
**                  client sessions to be hosted in the server.
** 007 EVL 20140528 Fix for displaying the labels for hidden widgets in a frame.  When the hidden
**                  widget become visible the label text should be drawn even later the widget
**                  will be hidden again.
*/

package com.goldencode.p2j.ui.client.widget;

import java.util.*;
 
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.layout.*;
import com.goldencode.p2j.util.*;

/**
 * Abstract widget container.
 */
public abstract class AbstractContainer
extends AbstractWidget
implements Container
{
   /** Stores context-local state variables. */
   private static final ContextLocal<WorkArea> work = 
   new ContextLocal<WorkArea>()
   {
      protected synchronized WorkArea initialValue()
      {
         return new WorkArea();
      }
   };

   /** Layout manager. */
   private LayoutManager layoutManager = null;

   /** Current (in focus) widget. */
   private Widget current = null;

   /** Container insets. */
   private Insets insets = new Insets(0, 0, 0, 0);

   /** Container size. */
   private Dimension size = new Dimension(1, 1);

   /** Widgets stored in the container. */
   private ArrayList<Widget>  widgets = new ArrayList<Widget>();

   /**
    * Constructor.
    */
   public AbstractContainer()
   {
   }

   /**
    * Add widget to container.
    *
    * @param   widget
    *          Widget to add.
    *
    * @return  <code>this</code> for fluent syntax support.
    */
   @Override
   public Container add(Widget widget)
   {
      widgets.add(widget);

      widget.setParent(this);

      return this;
   }

   /**
    * This method should be overridden in a container that allows clipping.
    *
    * @return  <code>true</code> if container clips its children.
    */
   @Override
   public boolean clipping()
   {
      return false;
   }

   /**
    * Get widget which is in focus.
    *
    * @return  focused widget.
    */
   @Override
   public Widget currentFocus()
   {
      if (current instanceof Container)
         return ((Container) current).currentFocus();

      return current;
   }

   /**
    * Get default focusable widget.
    *
    * @return  default focusable widget.
    */
   @Override
   public Widget defaultFocus()
   {
      Widget defFocus = null;

      for (Widget c : widgets)
      {
         if (c.focusTraversable())
         {
            defFocus = c;
            break;
         }
      }

      if (defFocus == null)
         throw new IllegalStateException("no focus-traversable widgets found");

      if (defFocus instanceof Container)
         return ((Container) defFocus).defaultFocus();

      return defFocus;
   }

   /**
    * Get widget dimension.
    *
    * @return  widget dimension.
    */
   @Override
   public Dimension dimension()
   {
      return new Dimension(width(), height());
   }

   /**
    * Lay out container.
    */
   @Override
   public void doLayout()
   {
      if (layoutManager != null)
         layoutManager.doLayout(this);
   }

   /**
    * Draw widget.
    */
   @Override
   public void draw()
   {
      if (!isVisible())
         return;

      Widget parent = parent();
      int lim = widgets.size();
      int from = 0;
      Widget c = null;

      OutputManager screen = screen();

      if (parent instanceof TitledWindow)
      {
         // new method for the top container
         // search down the widgets until all clipped widgets are
         // touched; from those, return the last found frame.
         screen.startScreenSearch();
         int lastFrom = lim - 1;
         boolean foundContainer = false;

         for (from = lim - 1; from >= 0; from--)
         {
            c = widgets.get(from);

            if (!c.isVisible())
            {
               continue;
            }

            if (!(c instanceof Container))
            {
               if (screen.markArea(c.bounds()) && screen.getInvalidChars() == 0)
               {
                  from = lastFrom;
                  break;
               }
               continue;
            }

            // in redirected mode, only the topmost frame is of interest
            if (screen.isRedirected())
               break;

            if (screen.markArea(c.bounds()))
            {
               lastFrom = from;
               foundContainer = true;
               if (screen.getInvalidChars() == 0)
               {
                  break;
               }
            }
         }

         if (from < 0)
         {
            if (foundContainer && screen.getInvalidChars() > 0)
            {
               // if there are clipped characters left and we found a
               // container, then drawing must start from the last found
               // position
               from = lastFrom;
            }
            else
            {
               from = 0;
            }
         }
      }

      // old method for nested containers
      for (int i = from; i < lim; i++)
      {
         c = widgets.get(i);

         if (c.isVisible())
         {
            if (screen.drawNeeded(c.bounds().translate(c.parent().screenLocation())))
               c.draw();
            // mark the label widget as "was visible" becaue we are going to show it
            if (c instanceof LabeledWidget)
            {
               ((LabeledWidget)c).setWasVisible();
            }
         }
      }
   }

   /**
    * Indicates the ending of the resize process. All repainting between
    * <code>startSizeChange</code> and <code>endSizeChange</code> calls is
    * suppressed.
    *
    * @param    repaint
    *           If <code>true</code> then the current bounding rectangle and
    *           the rectangle saved during <code>startSizeChange</code> call
    *           are repainted.
    */
   @Override
   public void endSizeChange(boolean repaint)
   {
      WorkArea wa = work.get();
      
      Rectangle savedSize = wa.savedSizes.pop();

      if (repaint)
      {
         if (savedSize != null)
            repaint(savedSize);
         
         repaint();
      }
   }

   /**
    * Returns the focused direct child Component.
    *
    * @return   The focused direct child Component.
    */
   @Override
   public Widget focus()
   {
      return current;
   }

   /**
    * Check if widget is focus traversable.
    *
    * @return  <code>true</code> if widget is focus traversable.
    */
   @Override
   public boolean focusTraversable()
   {
      if (!super.focusTraversable())
         return false;

      return hasFocusable(null);
   }

   /**
    * Get widget for provided index or <code>null</code> if no such component
    * exists.
    *
    * @param   index
    *          Widget index.
    *
    * @return  widget at specified index or <code>null</code> if index is out
    *          of range.
    */
   @Override
   public Widget get(int index)
   {
      if (index < 0 || index >= widgets.size())
         return null;

      return widgets.get(index);
   }

   /**
    * Get clipping rectangle.
    *
    * @return   clipping rectangle.
    */
   @Override
   public Rectangle getClipRect()
   {
      return new Rectangle(new Point(0, 0), dimension());
   }

   /**
    * Reports if there is at least 1 focusable widget that is not the given
    * widget.
    *
    * @param    exclude
    *           A widget to ignore or <code>null</code> if any focusable
    *           widget will do.
    *
    * @return   <code>true</code> if there is at least one other widget that
    *           can be focused.
    */
   @Override
   public boolean hasFocusable(Widget exclude)
   {
      for (Widget widget : widgets)
      {
         if (widget == exclude)
            continue;

         if (widget.focusTraversable())
            return true;
      }

      return false;
   }

   /**
    * Get widget height.
    *
    * @return  widget height.
    */
   @Override
   public int height()
   {
      return size().height;
   }

   /**
    * Get container insets.
    *
    * @return  container insets.
    */
   @Override
   public Insets insets()
   {
      return insets;
   }

   /**
    * Get widget minimum size.
    *
    * @return  widget minimum size.
    */
   @Override
   public Dimension minimumSize()
   {
      if (layoutManager == null)
         return size();

      return layoutManager.minimumSize(this);
   }

   /**
    * Mode specified widget to the bottom of the widget list.
    *
    * @param   widget
    *          Widget to mover.
    */
   @Override
   public void moveToBottom(Widget widget)
   {
      moveToZpos(widget, 0);
   }

   /**
    * Move specified widget to the top of widget list.
    *
    * @param   widget
    *          Widget to mover.
    */
   @Override
   public void moveToTop(Widget widget)
   {
      int size = widgets.size();

      if (size == 0)
         return;

      if (!widgets.contains(widget))
         return;

      if (widgets.get(size - 1).equals(widget))
         return;

      widgets.remove(widget);
      widgets.add(widget);
   }

   /**
    * Move widget to specified position in the widget list. For overlaying
    * widgets this is basically position in widget Z-order.
    *
    * @param   widget
    *          Widget to mover.
    * @param   pos
    *          Position in Z-order.
    */
   @Override
   public void moveToZpos(Widget widget, int pos)
   {
      if (pos < widgets.size() && widgets.contains(widget))
      {
         int ndx = widgets.indexOf(widget);

         if (ndx == pos)
            return;

         if (ndx < pos)
            pos--;

         widgets.remove(widget);
         widgets.add(pos, widget);
      }
   }

   /**
    * Select next focusable widget.
    */
   @Override
   public void nextFocus()
   {
      focusWorker(true);
   }

   /**
    * Select previous focusable widget.
    */
   @Override
   public void prevFocus()
   {
      focusWorker(false);
   }

   /**
    * Process key event.
    *
    * @param   event
    *          Key event to process.
    */
   @Override
   public void processKeyEvent(KeyInput event)
   {
      super.processKeyEvent(event);

      if (event.isConsumed())
         return;

      if (current != null && current != this)
         current.processKeyEvent(event);
   }

   /**
    * Remove specified widget from the widget list.
    *
    * @param   widget
    *          Widget to remove.
    */
   @Override
   public void remove(Widget widget)
   {
      widgets.remove(widget);

      widget.setParent(null);

      if (current == widget) {
         current = null;
         current = currentFocus();
      }
   }

   /**
    * Request widget repainting.
    */
   @Override
   public void repaint()
   {
      WorkArea wa = work.get();
      
      if (wa.savedSizes.isEmpty())
         super.repaint();
   }

   /**
    * Replace specified widget with another with preserving widget Z-order.
    *
    * @param   oldWidget
    *          Old widget.
    * @param   newWidget
    *          New widget.
    */
   @Override
   public void replace(Widget oldWidget, Widget newWidget)
   {
      if (oldWidget != null)
      {
         widgets.remove(oldWidget);
         oldWidget.setParent(null);

         if (current == oldWidget)
            current = newWidget;
      }

      add(newWidget);
   }

   /**
    * Request focus.
    */
   @Override
   public void requestFocus()
   {
      Widget focus = currentFocus();

      if (focus == null)
      {
         focus = defaultFocus();
      }
      if (focus == this)
      {
         super.requestFocus();
      }
      else if (focus != null)
      {
         focus.requestFocus();
      }
   }

   /**
    * Set focus to specified widget.
    *
    * @param   focus
    *          Widget to set focus on.
    */
   @Override
   public void setFocus(Widget focus)
   {
      current = focus;
      if (parent() != null)
         parent().setFocus(this);
   }

   /**
    * Set container height.
    *
    * @param   height
    *          New container height.
    */
   @Override
   public void setHeight(int height)
   {
      size.height = height;
   }

   /**
    * Set container insets.
    *
    * @param   insets
    *          Insets to set.
    */
   @Override
   public void setInsets(Insets insets)
   {
      this.insets = insets;
   }

   /**
    * Set layout manager for the container.
    *
    * @param   manager
    *          New layout manager.
    */
   @Override
   public void setLayout(LayoutManager manager)
   {
      this.layoutManager = manager;
   }

   /**
    * Get active layout manager for the container.
    *
    * @return  Active layout manager.
    */
   @Override
   public LayoutManager getLayout()
   {
      return layoutManager;
   }

   /**
    * Set container size.
    *
    * @param   size
    *          Size to set.
    */
   @Override
   public void setSize(Dimension size)
   {
      setSize(size.width, size.height);
   }

   /**
    * Set container size.
    *
    * @param   width
    *          New width.
    * @param   height
    *          New height.
    */
   @Override
   public void setSize(int width, int height)
   {
      if (isVisible())
      {
         repaint();
      }

      size = new Dimension(width, height);

      if (isVisible())
      {
         repaint(); // post repainting of the new rectangle
      }
   }

   /**
    * Set container width.
    *
    * @param   width
    *          Width to set.
    */
   @Override
   public void setWidth(int width)
   {
      size.width = width;
   }

   /**
    * Indicates the beginning of the resize process. All repainting between
    * <code>startSizeChange</code> and <code>endSizeChange</code> calls is
    * suppressed.
    *
    * @param    saveOldRect
    *           If <code>true</code> then the current bounding rectangle is
    *           saved for future repainting.
    */
   @Override
   public void startSizeChange(boolean saveOldRect)
   {
      WorkArea wa = work.get();
      
      if (saveOldRect)
      {
         wa.savedSizes.push(bounds());
      }
      else
      {
         wa.savedSizes.push(null);
      }
   }

   /**
    * Get all widgets stored in the container.
    *
    * @return  array of stored widgets.
    */
   @Override
   public Widget[] widgets()
   {
      return widgets.toArray(new Widget[widgets.size()]);
   }

   /**
    * Get widget width.
    *
    * @return  widget width.
    */
   @Override
   public int width()
   {
      return size().width;
   }

   /**
    * Add widget directly to the list without triggering repaint or other
    * operations.
    *
    * @param   widget
    */
   protected void addDirectly(Widget widget)
   {
      widgets.add(widget);
   }

   /**
    * Check if specified widget belongs to this container.
    *
    * @param   widget
    *          Widget to check.
    *
    * @return  <code>true</code> if widget belongs to container.
    */
   protected boolean belongsTo(Widget widget)
   {
      return widgets.contains(widget);
   }

   /**
    * Get container size.
    *
    * @return  container size.
    */
   //TODO: for some reason container uses separate method instead of dimension(),
   //      requires deeper investigation
   protected Dimension size()
   {
      return size;
   }

   /**
    * Set focus to first focusable widget.
    */
   @Override
   public void firstFocus()
   {
      Widget widget = locateFocusTraversable(widgets);

      if (widget instanceof Container)
         ((Container) widget).firstFocus();

      current = widget;
   }

   /**
    * Set focus to last focusable widget.
    */
   @Override
   public void lastFocus()
   {
      Widget widget = locateFocusTraversable(Iterables.reverse(widgets));

      if (widget instanceof Container)
         ((Container) widget).lastFocus();

      current = widget;
   }

   /**
    * Helper method for traversing widgets in the container.
    *
    * @param   iter
    *          Iterable for the contained widgets.
    *
    * @return  Found widget or <code>null</code> if no matches found.
    */
   private Widget locateFocusTraversable(Iterable<Widget> iter)
   {
      for (Widget widget : iter)
      {
         if (widget.focusTraversable())
            return widget;
      }

      return null;
   }

   /**
    * Search for the next or previous focusable widget in the container and
    * set focus on it.
    *
    * @param   direction
    *          Widget search direction.
    */
   private void focusWorker(boolean direction)
   {
      Iterable<Widget> iter = (direction) ?
                                     Iterables.directFrom(widgets, current) :
                                     Iterables.reverseFrom(widgets, current);

      if (iter == null)
         throw new IllegalStateException("Focused widget does not belong to container");

      for (Widget widget : iter)
      {
         if (checkWidget(widget, direction))
            return;
      }

      if (parent() != null)
      {
         if(direction)
            parent().nextFocus();
         else
            parent().prevFocus();
         return;
      }

      iter = (direction) ? widgets : Iterables.reverse(widgets);

      for (Widget widget : iter)
      {
         if (widget == current)
         {
            current = null;
            return;
         }

         if (checkWidget(widget, direction))
            return;
      }
   }

   /**
    * Check potential focus candidate and set focus on it if it matches
    * criteria.
    *
    * @param   widget
    *          Candidate widget.
    * @param   direction
    *          Search direction.
    *
    * @return  <code>true</code> if widget was focused and <code>false</code>
    *          otherwise.
    */
   private boolean checkWidget(Widget widget, boolean direction)
   {
      if (!widget.focusTraversable())
         return false;

      if (widget instanceof Container)
      {
         if (direction)
            ((Container) widget).firstFocus();
         else
            ((Container) widget).lastFocus();
      }

      widget.requestFocus();

      return true;
   }
   
   /** 
    * Stores global data relating to the state of the current context.
    */
   private static class WorkArea
   {  
      /** Container saved sizes stack. */       
      private Stack<Rectangle> savedSizes = new Stack<Rectangle>();
   }
}
